from ..tasks import DeployComponentTasks
from ..settings import DeployComponentSettings


class PythonComponentSettings(DeployComponentSettings):
    NAME = 'python'
    DEFAULT = {
        'root': 'python',
        'version': '3.7.3',
        'build': {
            'fast': True,
            'path': '/tmp/python'
        }
    }

    @property
    def version(self):
        return self._data['version']

    @property
    def fast_build(self):
        return self._data['build']['fast']

    @property
    def build_path(self):
        return self._data['build']['path']

    @property
    def site_packages_abs(self):
        return (
            self.root_abs / 'lib' /
            f'python{self.version[:3]}' / 'site-packages'
        )


class PythonComponentTasks(DeployComponentTasks):
    SETTINGS = PythonComponentSettings()

    async def pip(self, command: str):
        return await self.bin(f'pip3 {command}')

    async def run(self, command: str):
        return await self.bin(f'python3 {command}')

    async def bin(self, command: str):
        return await self._run(str(self.settings.root_abs / 'bin' / command))

    async def shell(self):
        return await self._run(
            str(self.settings.root_abs / 'bin' / 'python3'),
            interactive=True
        )

    async def build(self):
        version = self.settings.version
        fast_build = self.settings.fast_build
        build_path = self.settings.build_path

        await self._apt_install(
            'make', 'build-essential', 'libssl-dev', 'zlib1g-dev',
            'libbz2-dev', 'libreadline-dev', 'libsqlite3-dev', 'wget', 'curl',
            'llvm', 'libncurses5-dev', 'libncursesw5-dev', 'xz-utils',
            'tk-dev', 'tcl-dev', 'libffi-dev', 'wget'
        )
        await self._mkdir(self.settings.root, delete=True)

        async with self._cd(build_path, temporary=True):
            url = 'https://www.python.org/ftp/' \
                f'python/{version}/Python-{version}.tar.xz'
            await self._download_and_unpack(url)

            await self._run(
                './configure --prefix={0} '
                '--enable-loadable-sqlite-extensions --enable-shared '
                '--with-system-expat --enable-optimizations '
                'LDFLAGS="-L{0}/extlib/lib -Wl,--rpath={0}/lib '
                '-Wl,--rpath={0}/extlib/lib" '
                'CPPFLAGS="-I{0}/extlib/include"'.format(
                    self.settings.root_abs
                )
            )
            await self._run('make -j$(nproc) {}'.format(
                'build_all' if fast_build else 'build'))
            await self._run('make install > /dev/null')

        await self.pip('install wheel')
        await self.pip('install -U pip')
        await self.pip('install -U setuptools')



from unv.utils.os import get_homepath
from unv.utils.tasks import register, Tasks

from ..settings import SETTINGS


class VagrantTasks(Tasks):
    NAMESPACE = 'vagrant'

    @register
    async def setup(self):
        await self._local('vagrant destroy -f')
        await self._local('vagrant up')
        await self._update_local_known_hosts()

    async def _update_local_known_hosts(self):
        ips = [host['public_ip'] for _, host in SETTINGS.get_hosts()]
        known_hosts = get_homepath() / '.ssh' / 'known_hosts'

        if known_hosts.exists():
            with known_hosts.open('r+') as f:
                hosts = f.readlines()
                f.seek(0)
                for host in hosts:
                    if any(ip in host for ip in ips):
                        continue
                    f.write(host)
                f.truncate()

        for ip in ips:
            await self._local(f'ssh-keyscan {ip} >> {known_hosts}')



import jinja2

from pathlib import Path

from unv.deploy.settings import SETTINGS, DeployComponentSettings
from unv.deploy.tasks import DeployComponentTasks, register
from unv.deploy.components.systemd import SystemdTasksMixin


class IPtablesComponentSettings(DeployComponentSettings):
    NAME = 'iptables'
    DEFAULT = {
        'bin': '/sbin/iptables-restore',
        'user': 'root',
        'rules': {
            'template': 'ipv4.rules',
            'name': 'ipv4.rules'
        },
        'systemd': {
            'template': 'app.service',
            'name': 'iptables.service',
            'boot': True,
            'instances': {'count': 1}
        }
    }

    @property
    def rules_template(self):
        return self.local_root / self._data['rules']['template']

    @property
    def rules(self):
        return Path('/etc') / self._data['rules']['name']

    @property
    def bin(self):
        return f"{self._data['bin']} {self.rules}"


class IPtablesDeployTasks(DeployComponentTasks, SystemdTasksMixin):
    SETTINGS = IPtablesComponentSettings()

    @register
    async def sync(self):
        context = {
            'get_hosts': SETTINGS.get_hosts,
            'components': SETTINGS.get_components(self.public_ip)
        }
        rendered = []
        for task in self.get_all_deploy_tasks():
            get_template = getattr(task, 'get_iptables_template', None)
            if get_template is not None:
                template = jinja2.Template(
                    await get_template(), enable_async=True)
                context['deploy'] = task
                context['iptables_deploy'] = self
                rendered.append(await template.render_async(context))
                context.pop('deploy')
        context['components_templates'] = "\n".join([
            line.strip() for line in rendered
        ])

        await self._upload_template(
            self.settings.rules_template, self.settings.rules, context)
        await self._sync_systemd_units()

    @register
    async def setup(self):
        await self.sync()
        await self.start()



import copy
import inspect

from pathlib import Path

from unv.utils.collections import update_dict_recur
from unv.app.core import ComponentSettings


class DeploySettings(ComponentSettings):
    KEY = 'deploy'
    SCHEMA = {
        'hosts': {
            'type': 'dict',
            'keyschema': {'type': 'string'},
            'valueschema': {
                'type': 'dict',
                'schema': {
                    'public_ip': {'type': 'string', 'required': True},
                    'private_ip': {'type': 'string'},
                    'port': {'type': 'integer', 'required': True},
                    'provider': {'type': 'string', 'required': False},
                    'components': {
                        'type': 'list',
                        'schema': {'type': 'string'}
                    }
                }
            }
        },
        'components': {'allow_unknown': True}
    }
    DEFAULT = {
        'hosts': {},
        'components': {},
    }

    def get_hosts(self, component=''):
        for key, value in self._data['hosts'].items():
            if component in value.get('components', []) or not component:
                yield key, value

    def get_components(self, public_ip):
        for value in self._data['hosts'].values():
            if value['public_ip'] == public_ip:
                return value['components']
        return []

    def get_component_user(self, name):
        return self._data.get(name, {}).get('user', name)

    def get_component_settings(self, name):
        return self._data['components'].get(name, {})


SETTINGS = DeploySettings()


class DeployComponentSettings:
    def __init__(self, settings=None, root=None):
        if settings is None:
            settings = SETTINGS.get_component_settings(self.__class__.NAME)
        # TODO: add schema validation
        self._data = update_dict_recur(
            copy.deepcopy(self.__class__.DEFAULT), settings)
        self.local_root = root or Path(inspect.getfile(self.__class__)).parent

    @property
    def user(self):
        return self._data.get('user', self.__class__.NAME)

    @property
    def enabled(self):
        if 'enabled' in self._data:
            return self._data['enabled']
        for _, host in SETTINGS.get_hosts():
            if self.__class__.NAME in host['components']:
                return True
        return False

    @property
    def home(self):
        return Path('~')

    @property
    def home_abs(self):
        return Path('/', 'home', self.user)

    @property
    def systemd(self):
        return self._data.get('systemd', {})

    @property
    def root(self):
        return self.home / self._data['root']

    @property
    def root_abs(self):
        return self.home_abs / self._data['root']



from pathlib import Path

from watchgod import awatch

from ...tasks import DeployComponentTasks, nohost, register, onehost
from ...settings import SETTINGS, DeployComponentSettings

from ..python import PythonComponentTasks, PythonComponentSettings
from ..systemd import SystemdTasksMixin


class AppComponentSettings(DeployComponentSettings):
    NAME = 'app'
    DEFAULT = {
        'bin': 'app',
        'instance': 1,
        'settings': 'secure.production',
        'systemd': {
            'template': 'app.service',
            'name': 'app_{instance}.service',
            'boot': True,
            'instances': {'count': 1},
            'context': {
                'limit_nofile': 2000,
                'description': "Application description",
            }
        },
        'watch': {
            'dir': './src',
            'exclude': ['__pycache__', '*.egg-info']
        }
    }

    @property
    def python(self):
        settings = self._data.get('python', {})
        settings['user'] = self.user
        return PythonComponentSettings(settings)

    @property
    def bin(self):
        return str(self.python.root_abs / 'bin' / self._data['bin'])

    @property
    def module(self):
        return self._data['settings']

    @property
    def instance(self):
        return self._data['instance']

    @property
    def watch_dir(self):
        return Path(self._data['watch']['dir'])

    @property
    def watch_exclude(self):
        return self._data['watch']['exclude']


class AppComponentTasks(DeployComponentTasks, SystemdTasksMixin):
    SETTINGS = AppComponentSettings()

    def __init__(self, manager, lock, user, host, settings=None):
        super().__init__(manager, lock, user, host, settings)
        self._python = PythonComponentTasks(
            manager, lock, user, host, self.settings.python)

    @register
    @nohost
    async def watch(self):
        directory = self.settings.watch_dir
        site_packages_abs = self.settings.python.site_packages_abs

        async for _ in awatch(directory):
            for _, host in SETTINGS.get_hosts(self.NAMESPACE):
                with self._set_user(self.settings.user), self._set_host(host):
                    for sub_dir in directory.iterdir():
                        if not sub_dir.is_dir():
                            continue
                        await self._rsync(
                            sub_dir, site_packages_abs / sub_dir.name,
                            self.settings.watch_exclude
                        )
                    await self.restart()

    @register
    async def build(self):
        await self._create_user()
        await self._python.build()

    @register
    @onehost
    async def shell(self):
        return await self._python.shell()

    @register
    async def sync(self):
        name = (await self._local('python setup.py --name')).strip()
        version = (await self._local('python setup.py --version')).strip()
        package = f'{name}-{version}.tar.gz'

        await self._local('pip install -e .')
        await self._local('python setup.py sdist bdist_wheel')
        await self._upload(Path('dist', package))
        await self._local('rm -rf ./build ./dist')
        await self._python.pip(f'install -I {package}')
        await self._rmrf(Path(package))
        await self._upload(Path('secure'))
        await self._sync_systemd_units()

    @register
    async def setup(self):
        await self.build()
        await self.sync()
        await self.start()



import asyncio
import logging
import functools
import contextlib

from pathlib import Path

import jinja2

from unv.utils.tasks import Tasks, TasksManager, TaskRunError, register

from .settings import SETTINGS, DeployComponentSettings


def as_root(func):
    """Task will run from root, sets to self.user."""

    @functools.wraps(func)
    async def wrapper(self, *args, **kwargs):
        with self._set_user('root'):
            return await func(self, *args, **kwargs)

    return wrapper


def nohost(task):
    task.__nohost__ = True
    return task


def onehost(task):
    task.__onehost__ = True
    return task


class DeployTasks(Tasks):
    def __init__(self, manager, lock, user, host):
        self.user = user
        self.public_ip = host['public_ip']
        self.private_ip = host['private_ip']
        self.port = host['port']

        self._current_prefix = ''
        self._logger = logging.getLogger(self.__class__.__name__)
        self._lock = lock

        super().__init__(manager)

    async def _calc_instances_count(
            self, count: int = 0, cpu_count_percent: int = 0):
        if cpu_count_percent:
            cpu_cores = int(await self._run('nproc --all'))
            cpu_cores = int(cpu_cores / 100.0 * cpu_count_percent)
            cpu_cores += count
            count = cpu_cores
        return count or 1

    def get_all_deploy_tasks(self):
        for task_class in self._manager.tasks.values():
            if issubclass(task_class, DeployTasks):
                yield task_class(self._manager, self._lock, self.user, {
                    'public_ip': self.public_ip,
                    'private_ip': self.private_ip,
                    'port': self.port
                })

    @contextlib.contextmanager
    def _set_user(self, user):
        old_user = self.user
        self.user = user
        yield self
        self.user = old_user

    @contextlib.contextmanager
    def _set_host(self, host):
        old_public_ip, old_private_ip, old_port =\
            self.public_ip, self.private_ip, self.port
        self.public_ip, self.private_ip, self.port =\
            host['public_ip'], host['private_ip'], host['port']
        yield self
        self.public_ip, self.private_ip, self.port =\
            old_public_ip, old_private_ip, old_port

    @contextlib.contextmanager
    def _prefix(self, command):
        old_prefix = self._current_prefix
        self._current_prefix = f'{self._current_prefix} {command} '
        yield
        self._current_prefix = old_prefix

    @contextlib.asynccontextmanager
    async def _cd(self, path: Path, temporary=False):
        if temporary:
            await self._mkdir(path, delete=True)
        with self._prefix(f'cd {path} &&'):
            yield
        if temporary:
            await self._rmrf(path)

    @as_root
    async def _sudo(self, command, strip=True):
        """Run command on server as root user."""
        return await self._run(command, strip)

    async def _create_user(self):
        """Create user if not exist and sync ssh keys."""
        user = self.user

        with self._set_user('root'):
            try:
                await self._run("id -u {}".format(user))
            except TaskRunError:
                await self._run(
                    "adduser --quiet --disabled-password"
                    " --gecos \"{0}\" {0}".format(user)
                )

                local_ssh_public_key = Path('~/.ssh/id_rsa.pub')
                local_ssh_public_key = local_ssh_public_key.expanduser()
                keys_path = Path(
                    '/', 'home' if user != 'root' else '',
                    user, '.ssh'
                )

                await self._mkdir(keys_path)
                await self._run(f'chown -hR {user} {keys_path}')
                await self._run('echo "{}" >> {}'.format(
                    local_ssh_public_key.read_text().strip(),
                    keys_path / 'authorized_keys'
                ))

    @as_root
    async def _apt_install(self, *packages):
        with self._prefix('DEBIAN_FRONTEND=noninteractive'):
            await self._run('apt-get update -y -q')
            await self._run('apt-get upgrade -y -q')
            await self._run(
                'apt-get install -y -q --no-install-recommends '
                '--no-install-suggests {}'.format(' '.join(packages))
            )

    async def _run(self, command, strip=True, interactive=False) -> str:
        self._logger.debug(
            f'run [{self.user}@{self.public_ip}:{self.port}] '
            f'{self._current_prefix}{command}'
        )
        interactive_flag = '-t' if interactive else ''
        response = await self._local(
            f"ssh {interactive_flag} -p {self.port} "
            f"{self.user}@{self.public_ip} "
            f"'{self._current_prefix}{command}'",
            interactive=interactive
        ) or ''
        if strip:
            response = response.strip()
        return response

    async def _rmrf(self, path: Path):
        await self._run(f'rm -rf {path}')

    async def _mkdir(self, path: Path, delete=False):
        if delete:
            await self._rmrf(path)
        await self._run(f'mkdir -p {path}')

    async def _upload(self, local_path: Path, path: Path = '~/'):
        await self._local(
            f'scp -r -P {self.port} {local_path} '
            f'{self.user}@{self.public_ip}:{path}'
        )

    async def _rsync(self, local_dir, root_dir, exclude=None):
        exclude = [f"--exclude '{path}'" for path in exclude or []]
        exclude = ' '.join(exclude)
        await self._local(
            f'rsync -rave ssh --delete {exclude} {local_dir}/ '
            f'{self.user}@{self.public_ip}:{root_dir}'
        )

    async def _upload_template(
            self, local_path: Path, path: Path, context: dict = None):
        context = context or {}
        context.setdefault('deploy', self)

        async with self._lock:
            render_path = Path(f'{local_path}.render')
            template = jinja2.Template(
                local_path.read_text(), enable_async=True)
            render_path.write_text(await template.render_async(context))
            try:
                await self._upload(render_path, path)
            finally:
                render_path.unlink()

    async def _download_and_unpack(self, url: str, dest_dir: Path = Path('.')):
        await self._run(f'wget -q {url}')
        archive = url.split('/')[-1]
        await self._run(f'tar xf {archive}')
        archive_dir = archive.split('.tar')[0]

        await self._mkdir(dest_dir)
        await self._run(f'mv {archive_dir}/* {dest_dir}')

        await self._rmrf(archive)
        await self._rmrf(archive_dir)

    @register
    @onehost
    async def ssh(self):
        return await self._run('bash', interactive=True)


class DeployComponentTasks(DeployTasks):
    SETTINGS = None

    def __init__(self, manager, lock, user, host, settings=None):
        settings = settings or self.__class__.SETTINGS
        if settings is None or not isinstance(settings, DeployComponentSettings):
            raise ValueError(
                "Provide correct 'SETTINGS' value "
                "shoult be an instance of class 'DeployComponentSettings' not "
                f"[{settings}] value and type {type(settings)}"
            )
        self.settings = settings
        super().__init__(manager, lock, self.settings.user, host)

    @classmethod
    def get_namespace(cls):
        return cls.SETTINGS.NAME


class DeployTasksManager(TasksManager):
    def run_task(self, task_class, name, args):
        method = getattr(task_class, name)
        is_nohost = getattr(method, '__nohost__', False)
        is_onehost = getattr(method, '__onehost__', False)

        if issubclass(task_class, DeployTasks):
            if is_nohost:
                user = '__nohost__'
                hosts = [
                    ('', {'public_ip': None, 'private_ip': None, 'port': 0})
                ]
            else:
                user, hosts = self._get_user_with_hosts(task_class)

            if is_onehost and len(hosts) > 1:
                hosts_per_index = []
                for index, (host_name, host) in enumerate(hosts):
                    hosts_per_index.append([host_name, host])
                    print(f" ({index}) - {host_name} [{host['public_ip']}]")
                chosen_index = int(input('Please choose host to run task: '))
                hosts = [hosts_per_index[chosen_index]]

            async def run():
                lock = asyncio.Lock()
                tasks = [
                    getattr(task_class(self, lock, user, host), name)
                    for _, host in hosts
                ]
                await asyncio.gather(*[task(*args) for task in tasks])

            asyncio.run(run())
        else:
            return super().run_task(task_class, name, args)

    def _get_user_with_hosts(self, task_class):
        name = task_class.get_namespace()
        return (
            SETTINGS.get_component_user(name),
            list(SETTINGS.get_hosts(name))
        )



from pathlib import Path

from unv.utils.tasks import register

from ...tasks import DeployComponentTasks
from ...settings import DeployComponentSettings

from ..systemd import SystemdTasksMixin


class NginxComponentSettings(DeployComponentSettings):
    NAME = 'nginx'
    DEFAULT = {
        'systemd': {
            'template': 'server.service',
            'name': 'nginx.service',
            'boot': True,
            'instances': {'count': 1}
        },
        'master': True,
        'root': 'app',
        'packages': {
            'nginx': 'http://nginx.org/download/nginx-1.16.0.tar.gz',
            'pcre': 'https://ftp.pcre.org/pub/pcre/pcre-8.42.tar.gz',
            'zlib': 'http://www.zlib.net/zlib-1.2.11.tar.gz',
            'openssl': 'https://www.openssl.org/source/openssl-1.1.1a.tar.gz'
        },
        'configs': {'server.conf': 'nginx.conf'},
        'connections': 1000,
        'workers': 1,
        'aio': 'on',
        'sendfile': 'on',
        'tcp_nopush': 'on',
        'tcp_nodelay': 'on',
        'keepalive_timeout': 60,
        'include': 'conf/apps/*.conf',
        'access_log': 'logs/access.log',
        'error_log': 'logs/error.log',
        'default_type': 'application/octet-stream',
        'iptables': {
            'v4': 'ipv4.rules'
        }
    }

    @property
    def build(self):
        return self.root / 'build'

    @property
    def packages(self):
        return self._data['packages']

    @property
    def configs(self):
        for template, name in self._data['configs'].items():
            if not template.startswith('/'):
                template = (self.local_root / template).resolve()
            yield Path(template), self.root / 'conf' / name

    @property
    def include(self):
        return self.root_abs / self._data['include']

    @property
    def access_log(self):
        return self.root_abs / self._data['access_log']

    @property
    def error_log(self):
        return self.root_abs / self._data['error_log']

    @property
    def default_type(self):
        return self._data['default_type']

    @property
    def aio(self):
        return self._data['aio']

    @property
    def sendfile(self):
        return self._data['sendfile']

    @property
    def tcp_nopush(self):
        return self._data['tcp_nopush']

    @property
    def tcp_nodelay(self):
        return self._data['tcp_nodelay']

    @property
    def keepalive_timeout(self):
        return self._data['keepalive_timeout']

    @property
    def workers(self):
        return self._data['workers']

    @property
    def connections(self):
        return self._data['connections']

    @property
    def master(self):
        return self._data['master']

    @property
    def iptables_v4_rules(self):
        return (self.local_root / self._data['iptables']['v4']).read_text()


class NginxComponentTasks(DeployComponentTasks, SystemdTasksMixin):
    SETTINGS = NginxComponentSettings()

    async def get_iptables_template(self):
        return self.settings.iptables_v4_rules

    @register
    async def build(self):
        if not self.settings.master:
            print('Nginx already builded on this host, just use nginx.sync')
            return

        await self._create_user()
        await self._mkdir(self.settings.include.parent)
        await self._apt_install(
            'build-essential', 'autotools-dev', 'libexpat-dev',
            'libgd-dev', 'libgeoip-dev', 'libluajit-5.1-dev',
            'libmhash-dev', 'libpam0g-dev', 'libperl-dev',
            'libxslt1-dev'
        )

        async with self._cd(self.settings.build, temporary=True):
            for package, url in self.settings.packages.items():
                await self._download_and_unpack(url, Path('.', package))

            async with self._cd('nginx'):
                await self._run(
                    f"./configure --prefix={self.settings.root_abs} "
                    f"--user='{self.user}' --group='{self.user}' "
                    "--with-pcre=../pcre "
                    "--with-pcre-jit --with-zlib=../zlib "
                    "--with-openssl=../openssl --with-http_ssl_module "
                    "--with-http_v2_module --with-threads "
                    "--with-file-aio"
                )
                await self._run('make')
                await self._run('make install')

    @register
    async def sync(self):
        for template, path in self.settings.configs:
            await self._upload_template(template, path)

        for task in self.get_all_deploy_tasks():
            get_configs = getattr(task, 'get_nginx_include_configs', None)
            if get_configs is not None:
                configs = await get_configs()
                for template, path in configs:
                    await self._upload_template(
                        template,
                        self.settings.root / self.settings.include.parent
                        / path, {'deploy': task, 'nginx_deploy': self}
                    )

        await self._sync_systemd_units()

    @register
    async def setup(self):
        await self.build()
        await self.sync()
        await self.start()



from pathlib import Path

from unv.utils.tasks import register

from ..tasks import as_root


class SystemdTasksMixin:
    async def _get_systemd_services(self):
        systemd = self.settings.systemd
        name = systemd['name']
        count = await self._calc_instances_count(**systemd['instances'])
        for instance in range(1, count + 1):
            service = systemd.copy()
            service['name'] = name.format(instance=instance)
            service['instance'] = instance
            yield service

    @as_root
    async def _sync_systemd_units(self):
        services = [service async for service in self._get_systemd_services()]
        for service in services:
            service_path = Path('/etc', 'systemd', 'system', service['name'])
            context = {'instance': service['instance']}.copy()
            context.update(service.get('context', {}))
            path = service['template']
            if not str(path).startswith('/'):
                path = (self.settings.local_root / service['template'])
                path = path.resolve()
            await self._upload_template(path, service_path, context)

        await self._run('systemctl daemon-reload')

        for service in services:
            if service['boot']:
                await self._run(f'systemctl enable {service["name"]}')

    async def _systemctl(self, command: str, display=False):
        results = {}
        async for service in self._get_systemd_services():
            if 'manage' in service and not service['manage']:
                continue

            result = await self._sudo(f'systemctl {command} {service["name"]}')
            results[service['name']] = result
        return results

    @register
    async def start(self):
        await self._systemctl('start')

    @register
    async def stop(self):
        await self._systemctl('stop')

    @register
    async def restart(self):
        await self._systemctl('restart')

    @register
    async def status(self):
        results = await self._systemctl('status')
        for service, result in results.items():
            print(result)
